from edu_rdm_integration.export_data.export import (
    BaseExportEntitiesData,
    ExportEntitiesData,
)
from edu_rdm_integration.management.general import (
    BaseExportEntityDataCommand,
)


class Command(BaseExportEntityDataCommand):
    """
    Команда для выгрузки данных сущностей РВД.
    """

    # flake8: noqa: A003
    help = 'Команда для выгрузки данных сущностей РВД за указанных период.'

    def _prepare_export_entities_data_class(self, *args, **kwargs) -> BaseExportEntitiesData:
        """Возвращает объект класса экспорта данных сущностей РВД."""
        return ExportEntitiesData(
            entities=kwargs.get('entities'),
            period_started_at=kwargs.get('period_started_at'),
            period_ended_at=kwargs.get('period_ended_at'),
        )



from abc import (
    ABCMeta,
)
from datetime import (
    datetime,
)
from typing import (
    TYPE_CHECKING,
)

from django.apps import (
    apps,
)

from educommon import (
    logger,
)
from educommon.audit_log.helpers import (
    get_models_table_ids,
)
from educommon.audit_log.models import (
    AuditLog,
)
from function_tools.runners import (
    BaseRunner,
)

from edu_rdm_integration.adapters.managers import (
    WebEduRunnerManager,
)
from edu_rdm_integration.consts import (
    DATETIME_FORMAT,
    LOGS_DELIMITER,
)
from edu_rdm_integration.models import (
    CollectingDataStageStatus,
    CollectingExportedDataStage,
)


if TYPE_CHECKING:
    from django.db.models import (
        Model,
    )


class BaseCollectingDataRunnerManager(WebEduRunnerManager, metaclass=ABCMeta):
    """
    Базовый менеджер ранеров функций сбора данных для интеграции с "Региональная витрина данных".
    """

    forced_run = True

    def __init__(
        self,
        logs_period_started_at: datetime,
        logs_period_ended_at: datetime,
        logs: list[AuditLog] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Логи для сущности
        self._logs = logs

        self._logs_period_started_at = logs_period_started_at
        self._logs_period_ended_at = logs_period_ended_at
        self._stage = CollectingExportedDataStage.objects.create(
            manager_id=self.uuid,
            logs_period_started_at=logs_period_started_at,
            logs_period_ended_at=logs_period_ended_at,
        )

        self._command_id = kwargs.get('command_id')

        logger.info(f'{LOGS_DELIMITER}logs period started at {self._logs_period_started_at.strftime(DATETIME_FORMAT)}')
        logger.info(f'{LOGS_DELIMITER}log period ended at {self._logs_period_ended_at.strftime(DATETIME_FORMAT)}')
        logger.info(f'{LOGS_DELIMITER}created {repr(self._stage)}')

    def _collect_runner_regional_data_mart_integration_entities(
        self,
        runner_class: type[BaseRunner],
        runner_regional_data_mart_integration_entities: list[str],
    ):
        """
        Собирает и возвращает список сущностей.
        """
        for runnable_class in runner_class._prepare_runnable_classes():
            if hasattr(runnable_class, '_prepare_runnable_classes'):
                self._collect_runner_regional_data_mart_integration_entities(
                    runner_class=runnable_class,
                    runner_regional_data_mart_integration_entities=runner_regional_data_mart_integration_entities,
                )

                continue

            if hasattr(runnable_class, 'entities'):
                entities = getattr(runnable_class, 'entities')

                runner_regional_data_mart_integration_entities.extend(entities)

    def _get_loggable_models(self) -> set['Model']:
        """
        Возвращает перечень моделей по которым собираются логи.
        """
        loggable_models = set()
        regional_data_mart_integration_entities = []

        self._collect_runner_regional_data_mart_integration_entities(
            self.runner_class,
            regional_data_mart_integration_entities,
        )
        for entity in regional_data_mart_integration_entities:
            loggable_models.update(entity.loggable_models)
            # TODO: EDUSCHL-20938 Произвести рефакторинг plugins_info.
            if hasattr(entity, 'plugins_info'):
                for app_name, app_contents in entity.plugins_info.items():
                    if not apps.is_installed(app_name):
                        continue

                    for app_content in app_contents:
                        app_label, model_name = app_content.split('.')
                        try:
                            model = apps.get_model(app_label=app_label, model_name=model_name)
                            loggable_models.add(model)
                        except (ValueError, LookupError):
                            continue

        return loggable_models

    def _collect_logs(self):
        """
        Сбор логов для дальнейшей обработки.
        """
        return AuditLog.objects.filter(
            time__gte=self._logs_period_started_at,
            time__lt=self._logs_period_ended_at,
            table_id__in=get_models_table_ids(self._get_loggable_models()),
        ).order_by('time').iterator()

    def _create_runner(self, *args, **kwargs):
        """
        Метод создания ранера.
        """
        collected_logs = self._logs or self._collect_logs()

        logger.info(f'{LOGS_DELIMITER}{self.__class__.__name__} start preparing logs records..')

        self._stage.status_id = CollectingDataStageStatus.IN_PROGRESS.key
        self._stage.save()

        super()._create_runner(
            *args,
            logs=collected_logs,
            stage=self._stage,
            is_force_fill_cache=False,
            **kwargs,
        )

        logger.info(f'{LOGS_DELIMITER}change status {repr(self._stage)}')

    def _start_runner(self, *args, **kwargs):
        """
        Ранер необходимо запустить с отложенным заполнением кешей, чтобы заполнение произошло перед запуском объекта.
        """
        super()._start_runner(*args, is_force_fill_cache=False, **kwargs)

    def _after_start_runner(self, *args, **kwargs):
        """
        Точка расширения поведения менеджера ранера после запуска ранера.
        """
        if self._runner.result.errors:
            self._stage.status_id = CollectingDataStageStatus.FAILED.key
        else:
            self._stage.status_id = CollectingDataStageStatus.FINISHED.key

        self._stage.save()

        logger.info(f'{LOGS_DELIMITER}change status {repr(self._stage)}')



from typing import (
    Optional,
)

from edu_rdm_integration.adapters.strategies import (
    WebEduSyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy,
)
from edu_rdm_integration.collect_data.calculated.base.functions import (
    BaseCollectingCalculatedExportedDataFunction,
)
from edu_rdm_integration.collect_data.calculated.base.helpers import (
    BaseCollectingCalculatedExportedDataFunctionHelper,
    BaseCollectingCalculatedExportedDataRunnerHelper,
)
from edu_rdm_integration.collect_data.calculated.base.managers import (
    BaseCollectingCalculatedExportedDataRunnerManager,
)
from edu_rdm_integration.collect_data.calculated.base.runners import (
    BaseCollectingCalculatedExportedDataRunner,
)
from edu_rdm_integration.collect_data.calculated.base.validators import (
    BaseCollectingCalculatedExportedDataFunctionValidator,
    BaseCollectingCalculatedExportedDataRunnerValidator,
)
from edu_rdm_integration.collect_data.non_calculated.base.caches import (
    BaseCollectingExportedDataFunctionCacheStorage,
    BaseCollectingExportedDataRunnerCacheStorage,
)
from edu_rdm_integration.collect_data.non_calculated.base.errors import (
    BaseCollectingExportedDataError,
)
from edu_rdm_integration.collect_data.non_calculated.base.presenters import (
    BaseCollectingExportedDataResultPresenter,
)
from edu_rdm_integration.collect_data.non_calculated.base.results import (
    BaseCollectingExportedDataFunctionResult,
    BaseCollectingExportedDataRunnerResult,
)
from edu_rdm_integration.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
)


class CollectingCalculatedExportedDataFunctionImplementationStrategy(
    WebEduSyncBaseRunnerLazySavingPredefinedQueueFunctionImplementationStrategy,
):
    """
    Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение.

    Используется для функций сбора расчетных данных для выгрузки в "Региональная витрина данных".
    """

    @classmethod
    def _prepare_uuid(cls):
        """
        Получение UUID класса. Используется при регистрации сущности в базе данных.

        Если ничего не возвращает, то регистрация в БД не будет произведена.
        """
        return '6f560b8e-b259-4cad-8cfc-be27bb71066f'

    @classmethod
    def _prepare_verbose_name(cls) -> str:
        """
        Полное наименование для дальнейшей регистрации и отображения пользователю.
        """
        return (
            'Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение '
            'функций сбора расчетных данных для выгрузки в "Региональная витрина данных"'
        )

    @classmethod
    def _prepare_tags(cls) -> list[str]:
        """
        Список тегов, по которым сущность можно будет осуществлять поиск.
        """
        return [REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA]

    @classmethod
    def _prepare_key(cls) -> str:
        """
        Возвращает уникальный идентификатор стратегии создания функции.
        """
        return 'REGIONAL_DATA_MART_INTEGRATION_COLLECTING_CALCULATED_DATA'

    @classmethod
    def _prepare_title(cls) -> str:
        """
        Возвращает название стратегии создания функции.
        """
        return (
            'Стратегия создания функции с отложенным сохранением и предустановленной очередью объектов на сохранение '
            'функций сбора расчетных данных для выгрузки в "Региональная витрина данных". Сохранение производится '
            'после удачной работы функции'
        )

    @classmethod
    def _prepare_function_template_name(cls) -> Optional[str]:
        """
        Формирование названия шаблона создания функции.
        """
        return 'function_collect_data_template'

    def _prepare_manager_class(self):
        """
        Устанавливает класс менеджера.
        """
        self._manager_class = BaseCollectingCalculatedExportedDataRunnerManager

    def _prepare_runner_class(self):
        """
        Устанавливает класс ранера.
        """
        self._runner_class = BaseCollectingCalculatedExportedDataRunner

    def _prepare_function_class(self):
        """
        Устанавливает класс Функции.
        """
        self._function_class = BaseCollectingCalculatedExportedDataFunction

    def _prepare_runner_helper_class(self):
        """
        Устанавливает класс помощника ранера.
        """
        self._runner_helper_class = BaseCollectingCalculatedExportedDataRunnerHelper

    def _prepare_function_helper_class(self):
        """
        Устанавливает класс помощника функции.
        """
        self._function_helper_class = BaseCollectingCalculatedExportedDataFunctionHelper

    def _prepare_runner_validator_class(self):
        """
        Устанавливает класс валидатора ранера.
        """
        self._runner_validator_class = BaseCollectingCalculatedExportedDataRunnerValidator

    def _prepare_function_validator_class(self):
        """
        Устанавливает класс валидатора Функции.
        """
        self._function_validator_class = BaseCollectingCalculatedExportedDataFunctionValidator

    def _prepare_runner_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей ранера.
        """
        self._runner_cache_storage_class = BaseCollectingExportedDataRunnerCacheStorage

    def _prepare_function_cache_storage_class(self):
        """
        Устанавливает класс хранилища кешей Функции.
        """
        self._function_cache_storage_class = BaseCollectingExportedDataFunctionCacheStorage

    def _prepare_error_class(self):
        """
        Устанавливает класс ошибки.
        """
        self._error_class = BaseCollectingExportedDataError

    def _prepare_runner_result_class(self):
        """
        Устанавливает класс результата.
        """
        self._runner_result_class = BaseCollectingExportedDataRunnerResult

    def _prepare_function_result_class(self):
        """
        Устанавливает класс результата.
        """
        self._function_result_class = BaseCollectingExportedDataFunctionResult

    def _prepare_result_presenter_class(self):
        """
        Устанавливает класс презентера результата.
        """
        self._result_presenter_class = BaseCollectingExportedDataResultPresenter



from abc import (
    ABCMeta,
)

from function_tools.caches import (
    CacheStorage,
    EntityCache,
    PatchedGlobalCacheStorage,
    PeriodicalEntityCache,
)


class WebEduEntityCache(EntityCache, metaclass=ABCMeta):
    """
    Базовый класс кеша объектов сущности продуктов Образования.
    """


class WebEduPeriodicalEntityCache(PeriodicalEntityCache, metaclass=ABCMeta):
    """
    Базовый класс периодического кеша продуктов Образования.

    Кеш создается для определенной модели с указанием двух дат, на которые
    должны быть собраны кеши актуальных объектов модели.
    """


class WebEduRunnerCacheStorage(CacheStorage, metaclass=ABCMeta):
    """
    Базовый класс кешей помощников ранеров функций продуктов Образования.
    """


class WebEduFunctionCacheStorage(CacheStorage, metaclass=ABCMeta):
    """
    Базовый класс кешей функций продуктов Образования.
    """


class WebEduFunctionPatchedGlobalCacheStorage(PatchedGlobalCacheStorage, metaclass=ABCMeta):
    """
    Базовый класс кешей функций продуктов Образования с возможностью патчинга кешем глобального хелпера.
    """



from edu_rdm_integration.collect_data.base.runners import (
    BaseCollectingDataRunner,
)
from edu_rdm_integration.collect_data.calculated.base.helpers import (
    BaseCollectingCalculatedExportedDataRunnerHelper,
)
from edu_rdm_integration.collect_data.calculated.base.results import (
    BaseCollectingCalculatedExportedDataRunnerResult,
)
from edu_rdm_integration.collect_data.calculated.base.validators import (
    BaseCollectingCalculatedExportedDataRunnerValidator,
)


class BaseCollectingCalculatedExportedDataRunner(BaseCollectingDataRunner):
    """
    Базовый класс ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_helper_class(self) -> type[BaseCollectingCalculatedExportedDataRunnerHelper]:
        """
        Возвращает класс помощника ранера функции.
        """
        return BaseCollectingCalculatedExportedDataRunnerHelper

    def _prepare_validator_class(self) -> type[BaseCollectingCalculatedExportedDataRunnerValidator]:
        """
        Возвращает класс валидатора ранера функции.
        """
        return BaseCollectingCalculatedExportedDataRunnerValidator

    def _prepare_result_class(self) -> type[BaseCollectingCalculatedExportedDataRunnerResult]:
        """
        Возвращает класс результата ранера функции.
        """
        return BaseCollectingCalculatedExportedDataRunnerResult



from pathlib import (
    Path,
)
from typing import (
    Iterable,
)

from django.conf import (
    settings,
)
from django.db.models import (
    F,
    Q,
    Sum,
)
from django.db.transaction import (
    atomic,
)
from uploader_client.adapters import (
    adapter,
)

from educommon import (
    logger,
)

from edu_rdm_integration.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.export_data.base.consts import (
    OPERATIONS_METHODS_MAP,
    OPERATIONS_URLS_MAP,
)
from edu_rdm_integration.export_data.base.requests import (
    RegionalDataMartEntityRequest,
)
from edu_rdm_integration.export_data.consts import (
    TOTAL_ATTACHMENTS_SIZE_KEY,
)
from edu_rdm_integration.export_data.dataclasses import (
    UploadFile,
)
from edu_rdm_integration.export_data.queue import (
    Queue,
)
from edu_rdm_integration.models import (
    ExportingDataSubStage,
    ExportingDataSubStageAttachment,
    ExportingDataSubStageStatus,
    ExportingDataSubStageUploaderClientLog,
)
from edu_rdm_integration.redis_cache import (
    AbstractCache,
)


class ExportQueueSender:
    """Класс отправки данных в очередь РВД."""

    def __init__(self, data_cache: AbstractCache, queue: Queue):
        self.queue = queue
        self.cache = data_cache
        self.queue_total_file_size = self.cache.get(TOTAL_ATTACHMENTS_SIZE_KEY) or 0

    @staticmethod
    def get_exported_file_size() -> int:
        """Возвращает размер отправленных в РВД файлов.

        При расчете используются только файлы, отмеченные как отправленные, но не обработанные витриной.
        """
        sended_attachment_ids = ExportingDataSubStageUploaderClientLog.objects.filter(
            file_upload_status=FileUploadStatusEnum.IN_PROGRESS
        ).values_list('attachment_id', flat=True)
        file_size = ExportingDataSubStageAttachment.objects.filter(
            id__in=sended_attachment_ids
        ).aggregate(Sum('attachment_size'))['attachment_size__sum']

        return file_size or 0

    @staticmethod
    def _make_stage_filter(self) -> Q:
        """Формирование фильтра для выборки подэтапов."""
        return Q(status_id=ExportingDataSubStageStatus.READY_FOR_EXPORT.key)

    def get_sub_stages_attachments_to_export(self):
        """Выборка готовых к экспорту подэтапов."""
        return ExportingDataSubStage.objects.filter(
            self._make_stage_filter()
        ).annotate(
            attachment_id=F('exportingdatasubstageattachment__id'),
            attachment_name=F('exportingdatasubstageattachment__attachment'),
            attachment_size=F('exportingdatasubstageattachment__attachment_size'),
            operation=F('exportingdatasubstageattachment__operation'),
            entity=F('exportingdatasubstageentity__entity_id')
        ).order_by('started_at', 'operation').values(
            'id',
            'attachment_id',
            'attachment_name',
            'attachment_size',
            'operation',
            'entity',
        )

    def set_sub_stage_to_cache(self, sub_stage_id: int, entity_name: str, attachments: list[UploadFile]) -> bool:
        """Помещение подэтапа в очередь вместе с информацией по файлам и обновление счетчика объема файлов."""
        sub_stage_total_size = sum((attachment.attachment_size for attachment in attachments))
        if self.queue_total_file_size + sub_stage_total_size > settings.RDM_UPLOAD_QUEUE_MAX_SIZE:
            return False

        self.queue.enqueue(sub_stage_id, entity_name, attachments)
        self.queue_total_file_size += sub_stage_total_size
        # Обновим размер файлов в кеш
        self.cache.set(
            TOTAL_ATTACHMENTS_SIZE_KEY,
            self.queue_total_file_size,
            timeout=settings.RDM_REDIS_CACHE_TIMEOUT_SECONDS
        )

        logger.info(
            f'ExportedDataSubStage {sub_stage_id} {entity_name} added to the queue'
        )

        return True

    def run(self):
        """Запуск работы очереди."""
        if not self.queue_total_file_size:
            self.queue_total_file_size = self.get_exported_file_size()

        if self.queue_total_file_size < settings.RDM_UPLOAD_QUEUE_MAX_SIZE:
            stage_files = []
            prev_sub_stage = None
            entity = ''
            # Если размер очереди позволяет - то отправляем все файлы подэтапа в очередь - иначе прерываем процесс
            for stage_attachment in self.get_sub_stages_attachments_to_export():
                if prev_sub_stage != stage_attachment['id']:
                    if stage_files:
                        to_cache = self.set_sub_stage_to_cache(prev_sub_stage, entity, stage_files)

                        stage_files = []

                        if not to_cache:
                            break

                    prev_sub_stage = stage_attachment['id']

                if stage_attachment['attachment_size']:
                    stage_files.append(UploadFile(
                        stage_attachment['attachment_id'],
                        stage_attachment['attachment_name'],
                        stage_attachment['attachment_size'],
                        stage_attachment['operation']
                    ))
                    entity = stage_attachment['entity']

            # Обновляем общий объем очереди и закидываем последний элемент
            if stage_files:
                self.set_sub_stage_to_cache(prev_sub_stage, entity, stage_files)
        else:
            # Сохраняем объем отправленных файлов в кеш
            self.cache.set(
                TOTAL_ATTACHMENTS_SIZE_KEY,
                self.queue_total_file_size,
                timeout=settings.RDM_REDIS_CACHE_TIMEOUT_SECONDS
            )
            logger.warning(
                f'Total exported file size:  {self.queue_total_file_size} - queue is full!!!'
            )


class ExportEntityQueueSender(ExportQueueSender):
    """Класс отправки данных по конкретной сущности в очередь РВД."""

    def __init__(self, data_cache: AbstractCache, queue: Queue, entity: Iterable[str]):
        super().__init__(data_cache, queue)

        self.entity = entity

    def _make_stage_filter(self) -> Q:
        """Формирование фильтра для выборки подэтапов."""
        return Q(
            status_id=ExportingDataSubStageStatus.READY_FOR_EXPORT.key,
            exportingdatasubstageentity__entity_id__in=self.entity
        )


class WorkerSender:
    """Непосредственная отправка файлов."""

    def __init__(self, queue: Queue):
        self.queue = queue

    def send_files(self):
        """Отправка файлов."""
        sub_stages = self.queue.dequeue()
        for (sub_stage_id, entity_key), upload_files in sub_stages.items():
            sub_stage = ExportingDataSubStage.objects.filter(id=sub_stage_id).first()
            if not sub_stage:
                return

            result_to_save = []
            errors = []
            status = ExportingDataSubStageStatus.FINISHED.key
            for file in upload_files:
                if settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION:
                    logger.warning(
                        f'{LOGS_DELIMITER * 3}ATTENTION!!! REGIONAL DATA MART INTEGRATION REQUEST EMULATION ENABLED!'
                    )

                method = OPERATIONS_METHODS_MAP.get(file.operation)
                file_path = Path.joinpath(Path(settings.MEDIA_ROOT), file.attachment_name)

                try:
                    file_data = file_path.open('rb').read()
                except (OSError, IOError, FileNotFoundError) as error:
                    logger.error(f'Ошибка чтения файла {file_path} - {str(error)} ')
                    continue

                request = RegionalDataMartEntityRequest(
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    table_name=entity_key.lower(),
                    method=method,
                    operation=OPERATIONS_URLS_MAP.get(file.operation),
                    parameters={},
                    headers={
                        'Content-Type': 'text/csv',
                    },
                    files=[],
                    data=file_data,
                )

                result = adapter.send(request)

                request_id = ''
                if not result.error:
                    request_id = result.response.text

                file_upload_status = FileUploadStatusEnum.IN_PROGRESS if request_id else FileUploadStatusEnum.ERROR

                sub_stage_uploader_client_log = ExportingDataSubStageUploaderClientLog(
                    entry=result.log,
                    sub_stage_id=sub_stage_id,
                    attachment_id=file.attachment_id,
                    request_id=request_id,
                    file_upload_status=file_upload_status,
                    is_emulation=settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION,
                )
                result_to_save.append(sub_stage_uploader_client_log)
                if result.error:
                    logger.warning(
                        f'{result.error}\nrequest - "{result.log.request}"\nresponse - "{result.log.response}"'
                    )

                    errors.append(result.error)
                    status = ExportingDataSubStageStatus.FAILED.key
                else:
                    logger.info(
                        f'Response with {result.response.status_code} code and content {result.response.text}'
                    )

            # Сохраняем информацию об отправке файлов и убираем подэтап с файлами из очереди
            if result_to_save:
                with atomic():
                    ExportingDataSubStageUploaderClientLog.objects.bulk_create(result_to_save)
                    sub_stage.status_id = status
                    sub_stage.save()

            self.queue.delete_from_queue(sub_stage_id=sub_stage_id, entity_name=entity_key)
            logger.info(
                f'ExportedDataSubStage {sub_stage_id} {entity_key} sended from the queue'
            )

    def run(self):
        """Запуск воркера отправки."""
        self.send_files()



import inspect
import sys
from importlib import (
    import_module,
)
from operator import (
    attrgetter,
)
from typing import (
    Iterable,
)

from django.apps import (
    apps,
)
from django.db import (
    models,
)

from educommon.integration_entities.entities import (
    BaseEntity,
)
from m3_db_utils.mixins import (
    BaseEnumRegisterMixin,
)

from edu_rdm_integration.enum_register.mixins import (
    EntityEnumRegisterMixin,
    ModelEnumRegisterMixin,
)


def register_classes(classes: Iterable[type[BaseEnumRegisterMixin]]) -> None:
    """Вызывает метод регистрации в модель-перечисление
    у переданных классов.

    Args:
        classes: Классы, поддерживающие интерфейс регистрации.
    """
    classes = sorted(classes, key=attrgetter('order_number'))

    for enum_class in classes:
        enum_class.register()


def is_register_model(model: models.Model) -> bool:
    """Проверяет, является ли класс регистрируемой моделью."""
    return issubclass(model, ModelEnumRegisterMixin)


def is_register_entity(class_) -> bool:
    """Проверяет, является ли класс регистрируемой сущностью."""
    return (
        inspect.isclass(class_) and
        issubclass(class_, BaseEntity) and
        issubclass(class_, EntityEnumRegisterMixin)
    )


def register_models() -> None:
    """Регистрирует модели в RegionalDataMartModelEnum."""
    register_classes([
        m for m in apps.get_models() if is_register_model(m)
    ])


def register_entities(import_path: str) -> None:
    """Находит регистрируемые сущности в модуле по переданному пути и
    регистрирует в RegionalDataMartEntityEnum.

    Args:
        import_path: Путь до пакета, хранящего классы сущностей;
    """
    import_module(import_path)
    entities_module = sys.modules[import_path]

    register_classes([
        c[1] for c in inspect.getmembers(entities_module, is_register_entity)
    ])



from educommon.integration_entities.entities import (
    BaseEntity,
)

from edu_rdm_integration.adapters.caches import (
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)


class BaseExportDataRunnerCacheStorage(WebEduRunnerCacheStorage):
    """
    Базовый кеш помощников ранеров функций выгрузки данных для интеграции с "Региональная витрина данных".
    """


class BaseExportDataFunctionCacheStorage(WebEduFunctionCacheStorage):
    """
    Базовый кеш помощников функций выгрузки данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_entity_instances(self, model_ids, *args, **kwargs) -> list[BaseEntity]:
        """
        Формирование списка объектов сущностей для дальнейшей выгрузки.

        Необходимо переопределить логику формирования объектов у потомков.
        """
        return []

    def __init__(self, *args, model_ids, entities, **kwargs):
        super().__init__(*args, **kwargs)

        self._entity_enum_value = entities[0]
        self._model_ids = model_ids

    def _prepare(self, *args, **kwargs):
        """Заполнение кеша данными."""
        self.entity_instances = self._prepare_entity_instances(model_ids=self._model_ids)



# flake8: noqa
from django.db import (
    migrations,
)


def load_initial_data(apps, schema_editor):
    # Вместо модели UploadStatus создана модель-перечисление DataMartRequestStatus (см. миграцию 0004)
    pass


def delete_all_data(apps, schema_editor):
    """Удаление всех данных из модели UploadStatus при откате миграции."""
    apps.get_model('regional_data_mart_integration', 'UploadStatus').objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(load_initial_data, reverse_code=delete_all_data)
    ]



from abc import (
    ABCMeta,
)

from edu_rdm_integration.collect_data.base.managers import (
    BaseCollectingDataRunnerManager,
)
from edu_rdm_integration.collect_data.calculated.base.runners import (
    BaseCollectingCalculatedExportedDataRunner,
)


class BaseCollectingCalculatedExportedDataRunnerManager(BaseCollectingDataRunnerManager, metaclass=ABCMeta):
    """
    Менеджер ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    @classmethod
    def _prepare_runner_class(cls) -> type[BaseCollectingCalculatedExportedDataRunner]:
        """
        Возвращает класс ранера.
        """
        return BaseCollectingCalculatedExportedDataRunner



from edu_rdm_integration.collect_data.collect import (
    BaseCollectLatestModelsData,
    BaseCollectModelsData,
)
from edu_rdm_integration.management.general import (
    BaseCollectModelDataCommand,
)


class Command(BaseCollectModelDataCommand):
    """
    Команда для сбора на основе логов за период с последней сборки до указанной даты.
    """

    def _prepare_collect_models_data_class(self, *args, **kwargs) -> BaseCollectModelsData:
        return BaseCollectLatestModelsData(
            models=kwargs.get('models'),
            logs_period_started_at=kwargs.get('logs_period_started_at'),
            logs_period_ended_at=kwargs.get('logs_period_ended_at'),
        )



from abc import (
    ABCMeta,
)

from function_tools.helpers import (
    BaseFunctionHelper,
    BaseRunnerHelper,
)

from edu_rdm_integration.adapters.caches import (
    WebEduFunctionCacheStorage,
    WebEduRunnerCacheStorage,
)


class WebEduRunnerHelper(BaseRunnerHelper, metaclass=ABCMeta):
    """
    Базовый класс помощников ранеров функций продуктов Образования.
    """

    def _prepare_cache_class(self) -> type[WebEduRunnerCacheStorage]:
        """
        Возвращает класс кеша помощника ранера.
        """
        return WebEduRunnerCacheStorage


class WebEduFunctionHelper(BaseFunctionHelper, metaclass=ABCMeta):
    """
    Базовый класс помощников функций продуктов Образования.
    """

    def _prepare_cache_class(self) -> type[WebEduFunctionCacheStorage]:
        """
        Возвращает класс кеша помощника функции.
        """
        return WebEduFunctionCacheStorage



from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Optional,
    Union,
)

from django.contrib.contenttypes.models import (
    ContentType,
)
from django.db.models import (
    Q,
)
from django.db.models.constants import (
    LOOKUP_SEP,
)

from educommon.audit_log.utils import (
    get_model_by_table,
)
from educommon.integration_entities.consts import (
    LOG_OPERATION_MAP,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)

from edu_rdm_integration.collect_data.base.caches import (
    LogChange,
)
from edu_rdm_integration.helpers import (
    Graph,
)
from edu_rdm_integration.mapping import (
    MODEL_FIELDS_LOG_FILTER,
)
from edu_rdm_integration.utils import (
    build_related_model_graph,
)


if TYPE_CHECKING:
    from edu_rdm_integration.collect_data.base.caches import (
        IgnoreLogDependency,
    )


class ReformatLogsMixin:
    """Миксин для преобразования логов к удобному для работы виду в кешах помощников функций."""

    def _reformat_logs(self):
        """
        Производится преобразование логов к удобному для работы виду.

        Предполагается вложенные словари. На первом уровне ключом будет название
        модели, на втором идентификатор записи.
        """
        for log in self.raw_logs:
            model = get_model_by_table(log.table)._meta.label

            if getattr(self, '_log_only_models', None) and (model not in self._log_only_models):
                # Пропускаем, если модель не входит в список отслеживаемых
                continue

            operation = LOG_OPERATION_MAP[log.operation]

            if operation in EntityLogOperation.values:
                fields = log.transformed_data
            else:
                fields = {}

            # Если данных нет, то LogChange не формируем
            if not fields:
                continue

            log_change = LogChange(
                operation=operation,
                fields=fields,
            )

            if not self._filter_log(model, log_change):
                # Если модель не отслеживается, то запись лога не сохраняем
                continue

            if log_change.operation == EntityLogOperation.DELETE:
                self.logs[model][log.object_id] = [log_change, ]
            else:
                self.logs[model][log.object_id].append(log_change)

    @staticmethod
    def _filter_log(model: str, log_change: LogChange) -> bool:
        """
        Производится проверка изменений на отслеживаемые поля.
        """
        is_filtered = False

        if model in MODEL_FIELDS_LOG_FILTER[log_change.operation]:
            filter_fields = MODEL_FIELDS_LOG_FILTER[log_change.operation][model]
            if filter_fields:
                # Если заданы конкретные поля, которые должны отслеживать
                for field in log_change.fields:
                    if field in filter_fields:
                        # Достаточно, чтобы хотя бы одно поле попало под фильтр
                        is_filtered = True
                        break
            else:
                # Модель отслеживается, но перечень фильтруемых полей не задан,
                # значит фильтруем все поля модели
                is_filtered = True

        return is_filtered


class BaseIgnoreLogMixin:
    """Миксин для исключения логов не подлежащих обработке.

    Позволяет указать зависимость логов от состояния вышестоящих зависимых объектов.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._graph = Graph()
        self._dependency: Optional['IgnoreLogDependency'] = None

    @staticmethod
    def _add_lookup_prefix(q_filter: Q, lookup_path_args: list[str]) -> Q:
        """Функция принимает Q-объект и добавляет префикс к каждому указанному полю."""
        new_q = Q(_negated=q_filter.negated, _connector=q_filter.connector)

        for child in q_filter.children:
            if isinstance(child, Q):
                new_child = BaseIgnoreLogMixin._add_lookup_prefix(child, lookup_path_args)
            else:
                field_name, lookup_value = child

                if isinstance(field_name, str):
                    field_name = LOOKUP_SEP.join((*lookup_path_args, field_name))
                elif isinstance(field_name, tuple):
                    field_name = (LOOKUP_SEP.join((*lookup_path_args, field_name[0])),) + field_name[1:]

                new_child = (field_name, lookup_value)

            new_q.add(new_child, q_filter.connector)

        return new_q

    def _build_filter(
        self,
        lookup_path_args: list[str],
        filter_params: Union[tuple[Q], dict[str, Any]],
        is_valid_lookup: bool = False,
    ) -> Q:
        """Построение фильтра."""
        # Всегда строим OR, так как соответствие любому из параметров фильтра будет однозначно свидетельствовать, что
        # данные являются некорректными и должны быть исключены из последующей обработки
        q_filter = Q(_connector=Q.OR)

        if is_valid_lookup:
            if isinstance(filter_params, dict):
                for field_name, value in filter_params.items():
                    q_filter.add((LOOKUP_SEP.join((*lookup_path_args, field_name)), value), Q.OR)
            else:
                for filter_param in filter_params:
                    if isinstance(filter_param, Q):
                        q_filter |= self._add_lookup_prefix(filter_param, lookup_path_args)
                    else:
                        raise ValueError('Please, check `DependencyFilter.filters`, incorrect value!')

        return q_filter

    def _exclude_logs(self, model_label: str, object_model_ids: Iterable[int]):
        """Исключаем логи из обработки."""
        for object_id in object_model_ids:
            self.logs[model_label].pop(object_id)

    def _ignore_logs(self):
        """
        Исключение логов из обработки на основании описанных правил.
        """
        super()._ignore_logs()

        if not self._dependency:
            return

        model_label_separator = '.'

        for app_label, model_name in (model_label.split(model_label_separator) for model_label in self.logs):
            q_filter = Q(_connector=Q.OR)

            log_model = ContentType.objects.get(app_label=app_label.lower(), model=model_name.lower()).model_class()

            for dependency in self._dependency.dependency_filters:
                app_name, model_name = dependency.model_label.lower().split(model_label_separator)
                find_model = ContentType.objects.get(app_label=app_name, model=model_name).model_class()

                self._graph = build_related_model_graph(
                    self._graph, log_model, find_model, ignored_model_fields=self._dependency.ignore_model_fields
                )

                name_edges = self._graph.get_edges_between_vertices(
                    log_model._meta.model_name, find_model._meta.model_name
                )

                q_filter |= self._build_filter(
                    name_edges,
                    dependency.filters,
                    is_valid_lookup=bool(name_edges or find_model == log_model)
                )

            if q_filter:
                ignore_model_ids = log_model.objects.filter(
                    q_filter,
                    id__in=self.logs[log_model._meta.label],
                ).values_list('pk', flat=True)
                self._exclude_logs(log_model._meta.label, ignore_model_ids)



from typing import (
    Any,
    NamedTuple,
    Optional,
    Union,
)

from django.db.models import (
    Q,
)

from educommon.integration_entities.enums import (
    EntityLogOperation,
)


class LogChange(NamedTuple):
    """Операция и значения измененных полей из лога."""

    operation: EntityLogOperation
    fields: dict[str, Any]

    @property
    def is_create(self) -> bool:
        """Лог создания."""
        return self.operation == EntityLogOperation.CREATE

    @property
    def is_update(self) -> bool:
        """Лог изменения."""
        return self.operation == EntityLogOperation.UPDATE

    @property
    def is_delete(self) -> bool:
        """Лог удаления."""
        return self.operation == EntityLogOperation.DELETE


class DependencyFilter(NamedTuple):
    """Описание правила фильтрации данных модели."""

    model_label: str
    """
    Наименование модели от которой зависимы обрабатываемые логи.
    
    Строка должна соответствовать формату app_name.Model_name.
    """

    filters: Union[tuple[Q, ...], dict[str, Any]]
    """
    Описание полей и значений для описания фильтрации.
    """


class IgnoreLogDependency(NamedTuple):
    """Зависимость для исключения логов из последующей обработки."""

    dependency_filters: tuple[DependencyFilter, ...]
    """
    Описание фильтра применяемого к зависимости.
    """

    ignore_model_fields: Optional[dict[str, set[str]]] = None
    """
    Словарь, определяющий, какие поля моделей должны быть проигнорированы при построении графа связей между моделями.

    Ключом является наименование модели, а в множестве описываются наименования полей.
    """


# Generated by Django 2.2.28 on 2023-11-23 13:16

from django.db import (
    migrations,
)


FILE_UPLOAD_STATUS_FINISHED = 2

CREATE_INDEX_SQL = """
    CREATE INDEX rdm_exporting_data_sub_sta_file_upload_status ON public.rdm_exporting_data_sub_stage_uploader_client_log
    USING btree(file_upload_status)
    WHERE file_upload_status != %s;
"""

DROP_INDEX_SQL = """
    DROP INDEX IF EXISTS rdm_exporting_data_sub_sta_file_upload_status;
"""


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0002_init_data_uploadstatus'),
    ]

    operations = [
        migrations.RunSQL(
            [(CREATE_INDEX_SQL, [FILE_UPLOAD_STATUS_FINISHED])],
            DROP_INDEX_SQL,
        )
    ]



from abc import (
    ABCMeta,
)

from function_tools.errors import (
    BaseError,
)


class WebEduError(BaseError, metaclass=ABCMeta):
    """
    Базовый класс ошибок функций продуктов Образования.
    """
